import pickle, os
import keras
from keras import layers
from keras.models import load_model
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor, MLPClassifier
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import DotProduct, WhiteKernel
from tqdm import tqdm_notebook
import matplotlib.pyplot as plt


class Bridge:
    
    def __init__(self, config):
        self.bridge_type = config["Bridge"]["bridge_type"]
        self.hidden_layer_sizes = config["Bridge"]["hidden_layer_sizes"]
        self.activation = config["Bridge"]["activation"]
        self.solver = config["Bridge"]["solver"]
        self.alpha = config["Bridge"]["alpha"]
        self.batchsize = config[config["dataset_1"]]["model"]["train"]["batch_size"]
        self.n_epochs = config[config["dataset_1"]]["model"]["train"]["num_epochs"]
        self.lr = config["Bridge"]["lr"]
        self.lr_init = config["Bridge"]["lr_init"]
        self.max_iter = config["Bridge"]["max_iter"]
        self.shuffle = config["Bridge"]["shuffle"]
        self.random_state = config["Bridge"]["random_state"]
        self.tol = config["Bridge"]["tol"]
        self.verbose = config["Bridge"]["verbose"]
        self.validation_fraction = config["Bridge"]["validation_fraction"]
        self.restore_models = config["Bridge"]["restore_models"]
        self.input_shape = config[config["dataset_1"]]["latent_code_size"]
        self.output_shape = config["Bridge"]["subset"]
    
    def build_model(self):
        if self.bridge_type == "MLP":
            model = MLPRegressor(hidden_layer_sizes=(self.hidden_layer_sizes[0],),
                                    activation=self.activation,
                                    solver=self.solver,
                                    alpha=self.alpha,
                                    batch_size=self.batchsize,
                                    learning_rate=self.lr, 
                                    learning_rate_init=self.lr_init,
                                    max_iter=self.max_iter,
                                    shuffle=self.shuffle,
                                    random_state=self.random_state,
                                    tol=self.tol,
                                    verbose=self.verbose,
                                    validation_fraction=self.validation_fraction)
        elif self.bridge_type == "LR":
            model = LinearRegression()
        elif self.bridge_type == "GPR":
#             kernel = DotProduct() + WhiteKernel()
            model = GaussianProcessRegressor(random_state=42)
        return model
    
    def get_ae_bridge(self, input_shape, output_shape):
        if self.bridge_type == "DAE":
            input_layer = keras.Input(shape=(input_shape,))

            #### DAE-32
            encoded = layers.Dense(16, activation='relu')(input_layer)
            encoded = layers.Dense(8, activation='relu')(encoded)
#             encoded = layers.Dense(64, activation='relu')(encoded)

            decoded = layers.Dense(16, activation='relu')(encoded)
#             decoded = layers.Dense(512, activation='relu')(decoded)
            decoded = layers.Dense(output_shape)(decoded)
            num_decoder_layers = 2
        
        elif self.bridge_type == "LSTM":
            input_layer = keras.Input(shape=(input_shape, 1))

            x = layers.LSTM(64, activation='relu', return_sequences=True)(input_layer)
            x = layers.Dropout(0.2)(x)
    #             x = layers.LSTM(32, activation='relu', return_sequences=True)(x)
            encoded = layers.LSTM(32, activation='relu', return_sequences=False)(x)

            x = layers.RepeatVector(output_shape)(encoded)

#             x = layers.LSTM(32, activation='relu', return_sequences=True)(x)
    #             x = layers.LSTM(16, activation='relu', return_sequences=True)(x)
            x = layers.LSTM(64, activation='relu', return_sequences=True)(x)
            decoded = layers.TimeDistributed(layers.Dense(1))(x)
            num_decoder_layers = 3

        # Get Autoencoder
        autoencoder = keras.Model(input_layer, decoded)
        autoencoder.summary()
        # Get Encoder
        encoder = keras.Model(input_layer, encoded)
        # Get Decoder
        encoding_size = encoder.layers[-1].output_shape[1:]
        decoder_input = keras.Input(shape=(encoding_size))
        next_input = decoder_input
        for layer in autoencoder.layers[-num_decoder_layers:]:
            next_input = layer(next_input)
        decoder = keras.Model(inputs=decoder_input, outputs=next_input)
        autoencoder.compile(optimizer="adam", loss="mse")
        return autoencoder, encoder, decoder, encoding_size
    
    def train_bridge(self, bridge_ae, encoder, decoder, train_in, train_out,
                 test_in, test_out, n_epochs=100, batchsize=32, restore_models=False):
        if self.bridge_type == "LSTM":
            train_in = train_in.reshape(train_in.shape[0], train_in.shape[1], 1)
            train_out = train_out.reshape(train_out.shape[0], train_out.shape[1], 1)
            test_in = test_in.reshape(test_in.shape[0], test_in.shape[1], 1)
            test_out = test_out.reshape(test_out.shape[0], test_out.shape[1], 1)
        elif self.bridge_type == "DAE":
            train_in = train_in.reshape(train_in.shape[0], -1)
            train_out = train_out.reshape(train_out.shape[0], -1)
            test_in = test_in.reshape(test_in.shape[0], -1)
            test_out = test_out.reshape(test_out.shape[0], -1)
        print(train_in.shape, train_out.shape)
        bridge_history = bridge_ae.fit(train_in, train_out, epochs=n_epochs, 
                                       batch_size=batchsize, shuffle=True,
                                       validation_data=(test_in, test_out))
        plt.figure()
        plt.plot(bridge_history.history['loss'])
        plt.plot(bridge_history.history['val_loss'])
        plt.title('model loss')
        plt.ylabel('loss')
        plt.xlabel('epoch')
        plt.legend(['train', 'test'], loc='upper right')
        plt.show()
        return bridge_ae, encoder, decoder
    
    def train_models(self, latent_train_1, latent_train_2, latent_test_1, latent_test_2, n, eid, num_engines):
        models = []
        steps = int(n / num_engines)
        start = eid * steps
        for i in tqdm_notebook(range(0, steps, self.output_shape)):
            if not (self.bridge_type == "DAE" or self.bridge_type == "LSTM"):
                latent_train_2_col = latent_train_2[:, start+i]
                latent_test_2_col = latent_test_2[:, start+i]
                model = self.build_model()
                model.fit(latent_train_1, latent_train_2_col)
            else:
                latent_train_2_col = latent_train_2[:, start+i:start+i+self.output_shape]
                latent_test_2_col = latent_test_2[:, start+i:start+i+self.output_shape]
                model, bridge_encoder, bridge_decoder, bridge_encoding_size = self.get_ae_bridge(input_shape=self.input_shape,
                                                                                                 output_shape=self.output_shape)
                model, bridge_encoder, bridge_decoder = self.train_bridge(model, bridge_encoder, bridge_decoder, 
                                                                          latent_train_1, latent_train_2_col, 
                                                                          latent_test_1, latent_test_2_col, 
                                                                          n_epochs=self.n_epochs, batchsize=self.batchsize,
                                                                          restore_models=False)
            models.append(model)
            print("=============================================== Done: ", i)
        return models

    def store_models(self, fname, models):
        with open(fname, "wb") as f:
            for i in tqdm_notebook(range(len(models))):
                pickle.dump(models[i], f)
    
    def load_models(self, fname):
        models = []
        if os.path.isfile(fname):
            with open(fname, "rb") as f:
                while True:
                    try:
                        models.append(pickle.load(f))
                    except EOFError:
                        break
        return models